/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package sync4j.exchange.items.calendar.dao;

import java.io.IOException;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.StringTokenizer;

import sync4j.foundation.pdi.event.Reminder;
import sync4j.foundation.pdi.event.RecurrencePattern;

import sync4j.exchange.DataAccessException;
import sync4j.exchange.httptransport.WebDavHttpTransport;
import sync4j.exchange.items.calendar.model.Calendar;
import sync4j.exchange.items.common.dao.ItemDAO;
import sync4j.exchange.util.StringTools;
import sync4j.exchange.xml.XmlParseException;
import sync4j.exchange.xml.XmlParser;


/*
 * This class implements methods to access calendar data
 * in exchange server datastore
 *
 * @author Fabio Maggi @ Funambol
 * @version $Id: CalendarDAO.java,v 1.19 2005/07/26 16:03:42 nichele Exp $
 *
 **/
public class CalendarDAO extends ItemDAO {

    //---------------------------------------------------------------- Constants
    public static final String TAG_ALLDAYEVENT        = "alldayevent"       ;
    public static final String TAG_BODY               =
        "textdescription"                                                   ;
    public static final String TAG_BUSY_STATUS        = "busystatus"        ;
    public static final String TAG_DATEEND            = "dtend"             ;
    public static final String TAG_DATESTART          = "dtstart"           ;

    public static final String TAG_LOCATION           = "location"          ;
    public static final String TAG_RESPONSE           = "a:response"        ;
    public static final String TAG_SUBJECT            = "subject"           ;

    public static final String TAG_REMINDER_SET       = "reminderset"       ;
    public static final String TAG_REMINDER_OFFSET    = "reminderoffset"    ;

    public static final String TAG_REPLY_TIME         = "replytime"         ;

    public static final String TAG_SENSITIVITY        = "sensitivity"       ;

    public static final String TAG_RECURRENCE         = "rrule"             ;

    public static final String TAG_SEQUENCE           = "sequence"          ;

    public static final String TAG_MILEAGE            = "mileage"           ;
    public static final String TAG_IMPORTANCE         = "importance"        ;

    public static final String TAG_CATEGORIES         = "categories"        ;

    public static final String TAG_REPLUID            = "repluid"           ;
    public static final String TAG_LAST_MODIFIED      = "getlastmodified"   ;

    private static final String WEBDAV_MSG_SELECT_CALENDARS                  =
    "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n"                          +
    "<D:searchrequest xmlns:D =\"DAV:\">\n"                                  +
    "<D:sql>\n"                                                              +
    "Select "                                                                +
    "\"http://schemas.microsoft.com/repl/repl-uid\" AS " + TAG_REPLUID       + ", "  +
    "\"urn:schemas:calendar:location\" AS "              + TAG_LOCATION      + ", "  +
    "\"urn:schemas:httpmail:subject\" AS "               + TAG_SUBJECT       + ", "  +
    "\"urn:schemas:httpmail:textdescription\" AS "       + TAG_BODY          + ", "  +
    "\"urn:schemas:calendar:dtstart\" AS "               + TAG_DATESTART     + ", "  +
    "\"urn:schemas:calendar:dtend\" AS "                 + TAG_DATEEND       + ", "  +
    "\"urn:schemas:calendar:alldayevent\" AS "           + TAG_ALLDAYEVENT   + ", "  +
    "\"urn:schemas:calendar:busystatus\" AS "            + TAG_BUSY_STATUS   + ", "  +
    "\"urn:schemas:calendar:sequence\" AS "              + TAG_SEQUENCE      + ", "  +
    "\"urn:schemas:calendar:rrule\" AS "                 + TAG_RECURRENCE    + ", "  +
    "\"http://schemas.microsoft.com/mapi/reminderset\" AS " + TAG_REMINDER_SET      + ", "  +
    "\"urn:schemas:calendar:reminderoffset\" AS "           + TAG_REMINDER_OFFSET   + ", "  +
    "\"urn:schemas:calendar:replytime\" AS "                + TAG_REPLY_TIME        + ", "  +
    "\"http://schemas.microsoft.com/mapi/sensitivity\" AS " + TAG_SENSITIVITY       + ", "  +
    "\"http://schemas.microsoft.com/exchange/mileage\" AS " + TAG_MILEAGE           + ", "  +
    "\"http://schemas.microsoft.com/exchange/keywords-utf8\" AS " + TAG_CATEGORIES  + ", "  +
    "\"urn:schemas:httpmail:importance\" AS "               + TAG_IMPORTANCE        + ", "  +
    "\"DAV:getlastmodified\" AS "                        + TAG_LAST_MODIFIED        + ", "  +
    "\"DAV:isfolder\" AS isfolder "                                          +
    "FROM \"/{0}/{1}/{2}\"  "                                                +
    "{3}"                                                                    +
    "</D:sql>\n"                                                             +
    "</D:searchrequest>"                                                     ;

    private static final String FIELD_PREFIX = "urn:schemas:calendar:"  ;

    private static final String
                 WEBDAV_INIT_HM            = "HM:"                      ;
    private static final String
                 WEBDAV_INIT_HN            = "HN:"                      ;
    private static final String
                 WEBDAV_INIT_MAPI          = "MAPI:"                    ;
    private static final String
                 WEBDAV_INIT_EXCHANGE      = "EX:"                      ;

    private static final String
                 WEBDAV_TAG_ALLDAY         = WEBDAV_INIT_HM             +
                                             TAG_ALLDAYEVENT            ;
    private static final String
                 WEBDAV_TAG_BODY           = WEBDAV_INIT_HN             +
                                             TAG_BODY                   ;
    private static final String
                 WEBDAV_TAG_BUSYSTATUS     = WEBDAV_INIT_HM             +
                                             TAG_BUSY_STATUS            ;
    private static final String
                 WEBDAV_TAG_DATEEND        = WEBDAV_INIT_HM             +
                                             TAG_DATEEND                ;
    private static final String
                 WEBDAV_TAG_DATESTART      = WEBDAV_INIT_HM             +
                                             TAG_DATESTART              ;
    private static final String
                 WEBDAV_TAG_LOCATION       = WEBDAV_INIT_HM             +
                                             TAG_LOCATION               ;
    private static final String
                 WEBDAV_TAG_SUBJECT        = WEBDAV_INIT_HN             +
                                             TAG_SUBJECT                ;
    private static final String
                 WEBDAV_TAG_IMPORTANCE     = WEBDAV_INIT_HN             +
                                             TAG_IMPORTANCE             ;
    private static final String
                 WEBDAV_TAG_REMINDER_SET   = WEBDAV_INIT_MAPI           +
                                             TAG_REMINDER_SET           ;
    private static final String
                 WEBDAV_TAG_REMINDER_OFFSET = WEBDAV_INIT_HM            +
                                              TAG_REMINDER_OFFSET       ;
    private static final String
                 WEBDAV_TAG_REPLY_TIME      = WEBDAV_INIT_HM            +
                                              TAG_REPLY_TIME            ;
    private static final String
                 WEBDAV_TAG_SENSITIVITY     = WEBDAV_INIT_MAPI          +
                                              TAG_SENSITIVITY           ;
    private static final String
                 WEBDAV_TAG_RECURRENCE      = WEBDAV_INIT_HM            +
                                              TAG_RECURRENCE            ;
    private static final String
                 WEBDAV_TAG_SEQUENCE        = WEBDAV_INIT_HM            +
                                              TAG_SEQUENCE              ;
    private static final String
                 WEBDAV_TAG_MILEAGE         = WEBDAV_INIT_EXCHANGE      +
                                              TAG_MILEAGE               ;
    private static final String
                 WEBDAV_TAG_CATEGORIES      = WEBDAV_INIT_EXCHANGE      +
                                              TAG_CATEGORIES            ;
    private static final String
                 EVERY_WEEKDAY             = "MO, TU, WE, TH, FR"       ;

    //------------------------------------------------------------- Private data


    private WebDavHttpTransport  webDavHttp          = null ;

    //------------------------------------------------------------- Costructors

    public CalendarDAO(String  exchangeServerHost,
                       int     exchangeServerPort)
        throws DataAccessException {

        this.webDavHttp         = new WebDavHttpTransport(exchangeServerHost,
                                                          exchangeServerPort);
    }
    //------------------------------------------------------------- Public methods

    /**
     * insert / update calendar<br>
     *
     * @param c
     * @param username
     * @param credentials
     * @param exchangeFolder
     *
     * @return new / updated calendar
     *
     * @throws sync4j.exchange.util.DataAccessException
     **/
    public Calendar setCalendar(Calendar c              ,
                                String   username       ,
                                String   credentials    ,
                                String   exchangeFolder )
        throws DataAccessException {

        String       response                   = null  ;

        String       id                         = null  ;

        StringBuffer webDavCalendarMsg          = null  ;
        String       webDavHeaderMsg            = null  ;

        String       startDate                  = null  ;
        String       endDate                    = null  ;
        String       subject                    = null  ;
        String       body                       = null  ;
        String       location                   = null  ;
        Boolean      allDayB                    = null  ;
        String       allDay                     = null  ;
        String       busyStatus                 = null  ;

        String       priority                   = null  ;
        Integer      mileage                    = null  ;

        int          reminderMinutesBeforeStart = 0     ;
        boolean      reminderSet                = false ;
        Integer      replyTime                  = null  ;
        String       sensitivity                = null  ;
        String       sequence                   = null  ;

        String       categories                 = null  ;

        String       server                     = null  ;
        String       resource                   = null  ;

        String       recurrence                 = null  ;

        String       rrule                      = null  ;

        RecurrencePattern rp = c.getRecurrencePattern();

        server   = getServerFromExchangeFolder   (exchangeFolder );
        resource = getResourceFromExchangeFolder (exchangeFolder );

        rrule = c.getRrule().getPropertyValueAsString();

        if (rp != null){

            try {
                recurrence = getExchangeRecurrence(rp);
            } catch (Exception e) {
                throw new DataAccessException(e.getMessage());
            }

        }

        try {
            startDate = dateToWebDavTag (c.getDtStart ().getPropertyValueAsString());
            endDate   = dateToWebDavTag (c.getDtEnd   ().getPropertyValueAsString());
        } catch (Exception e) {
            throw new DataAccessException(e.getMessage());
        }

        if (c.getSummary() != null) {
            subject = c.getSummary().getPropertyValueAsString();
        }

        if (c.getDescription() != null) {
            body = c.getDescription().getPropertyValueAsString();
        }

        if (c.getLocation() != null) {
            location = c.getLocation().getPropertyValueAsString();
        }

        if (c.getCategories() != null) {
            categories = c.getCategories().getPropertyValueAsString();
        }

        allDayB = c.getAllDay();

        if (allDayB != null) {

            if (allDayB.booleanValue()) {
                allDay = "1";
            } else {
                allDay = "0";
            }
        }

        if (c.getPriority() != null) {
            priority = c.getPriority().getPropertyValueAsString();
        }

        mileage = c.getMileage();

        Reminder reminder = c.getReminder();
        if (reminder != null) {
            reminderMinutesBeforeStart = reminder.getMinutes();
            reminderSet = reminder.isActive();
        }

        replyTime = c.getReplyTime();

        if (c.getClassEvent() != null) {
            sensitivity = c.getClassEvent().getPropertyValueAsString();
        }

        if (c.getSequence() != null) {
            sequence = c.getSequence().getPropertyValueAsString();
        }

        //
        // must be implemented in PDI
        //
        busyStatus = "0";

        webDavCalendarMsg = new StringBuffer();

        webDavCalendarMsg.
            append
                ("<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n"                        +
                 "<D:propertyupdate xmlns:D=\"DAV:\" "                                  +
                 "xmlns:EX=\"http://schemas.microsoft.com/exchange/\" "                 +
                 "xmlns:HM=\"urn:schemas:calendar:\" "                                  +
                 "xmlns:MAPI=\"http://schemas.microsoft.com/mapi/\" "                   +
                 "xmlns:X=\"xml:\" "                                                    +
                 "xmlns:HN=\"urn:schemas:httpmail:\" "                                  +
                 "xmlns:dt=\"urn:uuid:c2f41010-65b3-11d1a29f-00aa00c14882/\" >\n"       +
                 "<D:set>\n"                                                            +
                 " <D:prop>\n"                                                          +
                 "  <D:contentclass>urn:content-classes:appointment</D:contentclass>\n" +
                 "  <EX:outlookmessageclass>IPM.Appointment</EX:outlookmessageclass>\n" );

        if (subject != null) {
            subject  = StringTools.escapeXml(subject);
            webDavCalendarMsg.append("<"                +
                                     WEBDAV_TAG_SUBJECT +
                                     ">"                +
                                     subject            +
                                     "</"               +
                                     WEBDAV_TAG_SUBJECT +
                                     ">\n"              );
        }

        if (body != null) {
            body  = StringTools.escapeXml(body);
            webDavCalendarMsg.append("<"             +
                                     WEBDAV_TAG_BODY +
                                     ">"             +
                                     body            +
                                     "</"            +
                                     WEBDAV_TAG_BODY +
                                     ">\n"           );
        }

        if (location != null) {
            location  = StringTools.escapeXml(location);
            webDavCalendarMsg.append("<"                 +
                                     WEBDAV_TAG_LOCATION +
                                     ">"                 +
                                     location            +
                                     "</"                +
                                     WEBDAV_TAG_LOCATION +
                                     ">\n"               );
        }

        webDavCalendarMsg.append("<"                  +
                                 WEBDAV_TAG_DATESTART +
                                 ">"                  +
                                 startDate            +
                                 "</"                 +
                                 WEBDAV_TAG_DATESTART +
                                 ">\n"                );

        webDavCalendarMsg.append("<"                +
                                 WEBDAV_TAG_DATEEND +
                                 ">"                +
                                 endDate            +
                                 "</"               +
                                 WEBDAV_TAG_DATEEND +
                                 ">\n"              );

        if (allDay != null) {
            webDavCalendarMsg.append("<"               +
                                     WEBDAV_TAG_ALLDAY +
                                     ">"               +
                                     allDay            +
                                     "</"              +
                                     WEBDAV_TAG_ALLDAY +
                                     ">\n"             );
        }

        if (busyStatus != null) {
            webDavCalendarMsg.append("<"                   +
                                     WEBDAV_TAG_BUSYSTATUS +
                                     ">"                   +
                                     busyStatus            +
                                     "</"                  +
                                     WEBDAV_TAG_BUSYSTATUS +
                                     ">\n"                 );
        }

        webDavCalendarMsg.append("<"                   +
                                 WEBDAV_TAG_IMPORTANCE +
                                 ">"                   +
                                 priority              +
                                 "</"                  +
                                 WEBDAV_TAG_IMPORTANCE +
                                 ">\n"                 );

        webDavCalendarMsg.append("<"                       +
                                 WEBDAV_TAG_REMINDER_SET   +
                                 ">"                       +
                                 (reminderSet ? "1" : "0") +
                                 "</"                      +
                                 WEBDAV_TAG_REMINDER_SET   +
                                 ">\n"                     );

        webDavCalendarMsg.append("<"                               +
                                 WEBDAV_TAG_REMINDER_OFFSET        +
                                 ">"                               +
                                 (reminderMinutesBeforeStart * 60) +
                                 "</"                              +
                                 WEBDAV_TAG_REMINDER_OFFSET        +
                                 ">\n"                             );

        if (replyTime != null) {
            webDavCalendarMsg.append("<"                   +
                                     WEBDAV_TAG_REPLY_TIME +
                                     ">"                   +
                                     replyTime.intValue()  +
                                     "</"                  +
                                     WEBDAV_TAG_REPLY_TIME +
                                     ">\n"                 );
        }

        if (sensitivity != null) {
            webDavCalendarMsg.append("<"                    +
                                     WEBDAV_TAG_SENSITIVITY +
                                     ">"                    +
                                     sensitivity            +
                                     "</"                   +
                                     WEBDAV_TAG_SENSITIVITY +
                                     ">\n"                  );
        }

        if (sequence != null) {
            webDavCalendarMsg.append("<"                 +
                                     WEBDAV_TAG_SEQUENCE +
                                     ">"                 +
                                     sequence            +
                                     "</"                +
                                     WEBDAV_TAG_SEQUENCE +
                                     ">\n"               );
        }

        if (mileage != null) {
            webDavCalendarMsg.append("<"                +
                                     WEBDAV_TAG_MILEAGE +
                                     ">"                +
                                     mileage.intValue() +
                                     "</"               +
                                     WEBDAV_TAG_MILEAGE +
                                     ">\n"              );
        }

        if (categories != null) {
            webDavCalendarMsg.append("<EX:keywords-utf8>");

            StringTokenizer st = new StringTokenizer(categories, ";");
            while (st.hasMoreTokens()) {
                webDavCalendarMsg.append("<X:v>"               +
                                         st.nextToken().trim() +
                                         "</X:v>\n"            );
            }

            webDavCalendarMsg.append("</EX:keywords-utf8>");

        }

        if (recurrence != null && recurrence.length() > 0) {

            webDavCalendarMsg.append("<"                    +
                                     WEBDAV_TAG_RECURRENCE  +
                                     " dt:dt=\"mv.string\"" +
                                     ">"                    );

            webDavCalendarMsg.append("<X:v>"      +
                                     recurrence   +
                                     "</X:v>\n"   );

            webDavCalendarMsg.append("</"                  +
                                     WEBDAV_TAG_RECURRENCE +
                                     ">\n"                 );

        }

        webDavCalendarMsg.
                append("<HM:timezoneid dt:dt=\"int\"></HM:timezoneid>\n");

        webDavCalendarMsg.
            append(" </D:prop>\n"        +
                   "</D:set>\n"          +
                   "</D:propertyupdate>" );

        webDavHeaderMsg
            = MessageFormat.format(ItemDAO.WEBDAV_HEADER_PROPPATCH,
                new Object[] {
                        "/"                 +
                        server              +
                        "/"                 +
                        username            +
                        "/"                 +
                        resource                 ,
                        c.getHref()
              }
        );

        try {
             response = this.webDavHttp.sendRequest(webDavHeaderMsg              ,
                                                    credentials                  ,
                                                    webDavCalendarMsg.toString() ,
                                                    FILE_ENCODING                );
        } catch (Exception e) {
            throw new DataAccessException("Error parsing calendar", e);
        }

        try {
            int s = getStatusFromResponse(response);
            checkResponseStatus(s);
        } catch (Exception e) {

            throw new DataAccessException("USER "                   +
                                          username                  +
                                          " URI "                   +
                                          exchangeFolder            +
                                          " set Exchange calendar"  ,
                                          e)                        ;

        }

        try {
            id = XmlParser.getRuidFromResponse(response);
        } catch (XmlParseException e) {
            throw new DataAccessException("Error getting resource id", e);
        }

        c.setId(id);

        return c;

    }

    /**
     * Delete a <i>Calendar</i> from Exchange server.
     *
     * @param calendar
     * @param username
     * @param credentials
     * @param exchangeFolder
     *
     * @throws sync4j.exchange.util.DataAccessException
     *
     */
    public void removeCalendar(Calendar calendar       ,
                               String   username       ,
                               String   credentials    ,
                               String   exchangeFolder )
        throws DataAccessException {

        String webDavCalendarMsg       = null ;

        String webDavHeaderMsg         = null ;

        String response                = null ;

        String     server              = null ;
        String     resource            = null ;

        server   = getServerFromExchangeFolder   (exchangeFolder );
        resource = getResourceFromExchangeFolder (exchangeFolder );

        webDavCalendarMsg       = ""                      ;


        webDavHeaderMsg =
            MessageFormat.format(ItemDAO.WEBDAV_HEADER_REMOVE,
            new Object[] {
                            "/"                      +
                            server                   +
                            "/"                      +
                            username                 +
                            "/"                      +
                            resource                   ,
                            calendar.getHref()

            }
        );


        try {
             response = this.webDavHttp.sendRequest(webDavHeaderMsg, credentials, webDavCalendarMsg, FILE_ENCODING);
         } catch (Exception e) {
            throw new DataAccessException("Error parsing calendar", e);
        }

        try {
            int s = getStatusFromResponse(response);
            checkResponseStatus(s);
        } catch (Exception e) {

            throw new DataAccessException("USER "                        +
                                          username                       +
                                          " URI "                        +
                                          exchangeFolder                 +
                                          " removing Exchange calendar"  ,
                                          e)                             ;
        }

    }

    /**
     * get calendar array from Exchange Server
     * @param username
     * @param credentials
     * @param ids
     * @param exchangeFolder
     *
     * @return array of find calendars
     *
     * @throws sync4j.exchange.util.DataAccessException
     **/
    public Calendar[] getCalendars(String     username       ,
                                   String     credentials    ,
                                   String[]   ids            ,
                                   String     exchangeFolder )
        throws DataAccessException {

        String clause = null ;

        clause = getClause(ids);

        return getCalendars(username, credentials, clause, exchangeFolder);
    }

    /**
     * get calendar array from Exchange Server
     *
     * @param username
     * @param credentials
     * @param fields
     * @param values
     * @param exchangeFolder
     * @return array of find calendars
     *
     * @throws sync4j.exchange.util.DataAccessException
     **/
    public Calendar[] getCalendars(String username,
                                   String credentials,
                                   String fields[],
                                   Object values[],
                                   String exchangeFolder) throws DataAccessException {

        String clause                 = null ;

        clause = getClause(fields, values);

        return getCalendars(username, credentials, clause, exchangeFolder);
    }

    /**
     * Create a <i>Calendar</i> object from WebDav msg.
     *
     * @param id
     * @param webDavMsg
     *
     * @return a newly create <i>Calendar</i> object initialized with the
     *         fields in the result set
     *
     */
    private Calendar[] getCalendarsFromWebDavMsg(String webDavMsg)
    throws DataAccessException {

        ArrayList   calendars  = null ;
        String  []  resps      = null ;
        String  []  msg        = null ;

        String      response   = null ;
        String      id         = null ;
        String      replUid    = null ;

        String      isFolder   = null ;

        int         count      = 0    ;

        msg = new String[] {webDavMsg};

        calendars = new ArrayList();

        try {

            resps = XmlParser.getXMLTag(msg, TAG_RESPONSE);

            for (int i = 0, l = resps.length; i < l; i++) {

                response =  resps[i]                                         ;
                replUid  =  XmlParser.getXMLTagValue (response, TAG_REPLUID) ;
                id       =  getIdFromReplUid (replUid)                       ;

                isFolder =  XmlParser.getXMLInitTagValue (response,
                                                          TAG_IS_FOLDER)     ;

                if(PROP_NO_FOLDER.equals(isFolder)) {
                    calendars.add(getCalendarFromResponseTag(id, response));
                    count ++;
                }

            }

        } catch (XmlParseException e) {
            throw new DataAccessException("Error parsing calendar",
                                          e);
        }

        return (Calendar[])calendars.toArray(new Calendar[count]);

    }

    /**
     * Create a <i>Calendar</i> object from WebDav msg.
     *
     * @param id
     * @param msgResponse
     *
     * @return a newly create <i>Calendar</i> object initialized with the
     *         fields in the result set
     *
     */
    private Calendar getCalendarFromResponseTag(String id, String msgResponse)
    throws DataAccessException {

        Calendar c                 = null ;
        String   lastUpdate        = null ;

        String   dateStart         = null ;
        String   dateEnd           = null ;
        String   subject           = null ;
        String   body              = null ;
        String   location          = null ;
        Boolean  allDayB           = null ;
        String   allDay            = null ;
        String   busyStatus        = null ;
        String   importance        = null ;
        String   mileage           = null ;
        String   reminderOffset    = null ;
        String   reminderSet       = null ;
        String   replyTime         = null ;
        String   sensitivity       = null ;
        String   sequence          = null ;

        String[]  categoriesList   = null ;
        String    categories       = null ;
        String[]  recurrenceList   = null ;
        String    recurrence       = null ;
        String    value            = null ;

        try {

            body       = XmlParser.getXMLTagValue
                            (msgResponse, TAG_BODY)          ;

            location   = XmlParser.getXMLTagValue
                            (msgResponse, TAG_LOCATION)      ;

            subject    = XmlParser.getXMLTagValue
                            (msgResponse, TAG_SUBJECT)       ;

            busyStatus = XmlParser.getXMLTagValue
                            (msgResponse, TAG_BUSY_STATUS)   ;

            dateStart  = XmlParser.getXMLInitTagValue
                            (msgResponse, TAG_DATESTART)     ;

            dateEnd    = XmlParser.getXMLInitTagValue
                            (msgResponse, TAG_DATEEND)       ;

            lastUpdate = XmlParser.getXMLInitTagValue
                            (msgResponse, TAG_LAST_MODIFIED) ;

            allDay     = XmlParser.getXMLInitTagValue
                            (msgResponse, TAG_ALLDAYEVENT)   ;

            importance = XmlParser.getXMLInitTagValue
                            (msgResponse, TAG_IMPORTANCE);

            mileage = XmlParser.getXMLInitTagValue
                            (msgResponse, TAG_MILEAGE);

            reminderOffset = XmlParser.getXMLInitTagValue
                            (msgResponse, TAG_REMINDER_OFFSET);

            reminderSet = XmlParser.getXMLInitTagValue
                            (msgResponse, TAG_REMINDER_SET);

            replyTime = XmlParser.getXMLInitTagValue
                           (msgResponse, TAG_REPLY_TIME);

            sensitivity = XmlParser.getXMLInitTagValue
                           (msgResponse, TAG_SENSITIVITY);

            sequence = XmlParser.getXMLInitTagValue
                           (msgResponse, TAG_SEQUENCE);

            categories =
                XmlParser.getXMLInitTagValue(msgResponse, TAG_CATEGORIES);


            if (categories != null) {
                categoriesList =
                    XmlParser.getXMLTag(new String[] {categories}, "c:v");
            }
            if (categoriesList != null) {
                StringBuffer sb = new StringBuffer();
                for (int i=0; i<categoriesList.length; i++) {
                    sb.append(categoriesList[i]).append("; ");
                }
                categories = sb.toString();
            }


            recurrence =
                XmlParser.getXMLInitTagValue(msgResponse, TAG_RECURRENCE);


            if (recurrence != null) {
                recurrence = recurrence.trim();
                recurrenceList =
                    XmlParser.getXMLTag(new String[] {recurrence}, "c:v");
            }
            if (recurrenceList != null) {

                StringBuffer sb = new StringBuffer();

                int l = recurrenceList.length;

                if (l == 1) {

                    sb.append(recurrenceList[0]);

                } else {

                    for (int i=0; i < l; i++) {

                        sb.append(recurrenceList[i]).append("; ");
                    }

                }

                recurrence = sb.toString();
            }

            c = new Calendar(id, XmlParser.webDavTagToDate(lastUpdate));

            c.getDtStart ().setPropertyValue(XmlParser.webDavTagToPDIDate(dateStart ));
            c.getDtEnd   ().setPropertyValue(XmlParser.webDavTagToPDIDate(dateEnd   ));

            if (body != null) {
                c.getDescription().setPropertyValue(body);
            }

            if (location != null) {
               c.getLocation().setPropertyValue(location);
            }

            if (subject != null) {
                c.getSummary().setPropertyValue(subject);
            }

            if (allDay != null) {
                if("1".equals(allDay)) {
                    c.setAllDay(new Boolean(true));
                } else if("0".equals(allDay)) {
                    c.setAllDay(new Boolean(false));
                }
            }

            if (importance != null) {
                c.getPriority().setPropertyValue(importance);
            }

            if (mileage != null) {
                c.setMileage(new Integer(mileage));
            }

            if (reminderOffset != null) {
                Reminder reminder = c.getReminder();
                if (reminder == null) {
                    reminder = new Reminder();
                    c.setReminder(reminder);
                }
                int minutes = ((new Integer(reminderOffset)).intValue()) / 60;
                reminder.setMinutes(minutes);
            }

            if (reminderSet != null) {
                Reminder reminder = c.getReminder();
                if (reminder == null) {
                    reminder = new Reminder();
                    c.setReminder(reminder);
                }
                if (reminderSet.equals("1")) {
                    reminder.setActive(true);
                } else {
                    reminder.setActive(false);
                }
            }

            if (replyTime != null) {
                c.setReplyTime(new Integer(replyTime));
            }

            if (sensitivity != null) {
                c.getClassEvent().setPropertyValue(sensitivity);
            }

            if (sequence != null) {
                c.getSequence().setPropertyValue(sequence);
            }

            if (categories != null) {
                c.getCategories().setPropertyValue(categories);
            }

            if (recurrence != null) {
                c.setRecurrencePattern(
                            getRecurrencePattern(recurrence,
                                XmlParser.webDavTagToPDIDate(dateStart )));
            }

            //
            // busyStatus must be implemented in PDI
            //

        } catch (Exception e) {
            throw new DataAccessException("Error parsing calendar item: ", e);
        }

        return c;

    }

    /**
     * Build a clause
     * to webdav select request
     * between ids item array
     *
     * @param ids
     * @return webdav select clause
     **/
    private static String getClause (String[] fields, Object[] values) {

        StringBuffer clause = new StringBuffer();

        int l = fields.length;

        if (l == 0) {
            return "";
        }

        clause.append("where (");

        StringBuffer tmp = null;
        String valueTmp  = null;

        for (int i = 0; i < l; i++) {
            tmp = new StringBuffer();

            valueTmp = "";

            if (TAG_DATESTART.equals(fields[i])) {
                tmp.append("\"").append(FIELD_PREFIX).append("dtstart\"=CAST(\"");

                if (values[i] != null) {
                    valueTmp = dateToWebDavTag((java.util.Date)values[i]);
                }
                tmp.append(valueTmp).append("\" as 'dateTime')");
            }

            if (TAG_DATEEND.equals(fields[i])) {
                tmp.append("\"").append(FIELD_PREFIX).append("dtend\"=CAST(\"");

                if (values[i] != null) {
                    valueTmp = dateToWebDavTag((java.util.Date)values[i]);
                }

                tmp.append(valueTmp).append("\" as 'dateTime')");
            }

            if (TAG_LOCATION.equals(fields[i])) {
                tmp.append("\"").append(FIELD_PREFIX).append("location\"");

                if (values[i] != null) {
                    tmp.append("='");
                    valueTmp = String.valueOf(values[i]);
                    valueTmp = valueTmp.replaceAll("'", "''");
                    valueTmp = StringTools.escapeXml(valueTmp);
                tmp.append(valueTmp).append("'");
                } else {
                    tmp.append(" is null ");
                }

            }

            if (i != 0 && tmp.length() != 0) {
                clause.append(" and ").append(tmp);
            } else if (tmp.length() != 0) {
                clause.append(tmp);
            }

        }

        clause.append(")\r\n");

        return clause.toString();

    }


    /**
     * get calendars from Exchange Server with the given clause
     *
     * @param username
     * @param credentials
     * @param clause
     * @param exchangeFolder
     * @return array of find calendars
     *
     * @throws sync4j.exchange.util.DataAccessException
     **/
    private Calendar[] getCalendars(String username,
                                    String credentials,
                                    String clause,
                                    String exchangeFolder) throws DataAccessException {

        Calendar[] exchangeCalendars = null;

        String response = null;
        String webDavCalendarMsg = null;

        String webDavHeaderMsg = null;

        String server = null;
        String resource = null;

        server = getServerFromExchangeFolder(exchangeFolder);
        resource = getResourceFromExchangeFolder(exchangeFolder);

        if (clause == null || clause.length() == 0 ) {
            clause = " WHERE ";
        } else {
            clause = clause + " AND ";
        }

        clause = clause +
                 "not (\"http://schemas.microsoft.com/mapi/id/{00062002-0000-0000-C000-000000000046}/0x8231\" != cast(\"0\" as int) " +
                 "and (\"urn:schemas:calendar:instancetype\" != 1 and \"urn:schemas:calendar:instancetype\" != 0))" ;

        webDavCalendarMsg =
                MessageFormat.format(WEBDAV_MSG_SELECT_CALENDARS,
                                     new Object[] {server, username, resource,
                                     clause});

        webDavHeaderMsg =
                MessageFormat.format(ItemDAO.WEBDAV_HEADER_SELECT,
                                     new Object[] {
                                         "/" +
                                         server +
                                         "/" +
                                         username +
                                         "/" +
                                         resource
                                     }
                );

        try {
            response = this.webDavHttp.sendRequest(webDavHeaderMsg, credentials,
                                                   webDavCalendarMsg, FILE_ENCODING);
        } catch (IOException e) {
            throw new DataAccessException(
                    "Error getting Exchange server response",
                    e);
        }

        try {
            int s = getStatusFromResponse(response);
            checkResponseStatus(s);
        } catch (Exception e) {

            throw new DataAccessException("USER " +
                                          username +
                                          " URI " +
                                          exchangeFolder +
                                          " getting exchange calendar",
                                          e);

        }

        exchangeCalendars = getCalendarsFromWebDavMsg(response);

        return exchangeCalendars;

    }

    private RecurrencePattern getRecurrencePattern(String recurrence          ,
                                                   String startDatePattern    )
    throws Exception {

        RecurrencePattern rp = null;

        short   recurrenceType = 0    ;
        short   dayOfWeekMask  = 0    ;
        short   dayOfMonth     = 0    ;
        short   monthOfYear    = 0    ;
        short   instance       = 0    ;

        int     interval       = 0    ;

        String  rt             = null ;
        String  value          = null ;
        String  days           = null ;
        String  endDatePattern = null ;

        boolean noEndDate      = true ;

        rt = recurrence.substring(recurrence.indexOf("FREQ=") + "FREQ=".length(),
                                  recurrence.indexOf(";")).trim() ;

        //
        // start date pattern
        //
        if (recurrence.indexOf("UNTIL") != -1) {
            noEndDate = false;
            int start = recurrence.indexOf("UNTIL=") + "UNTIL=".length();
            int end   = recurrence.indexOf(";", start);
            endDatePattern = recurrence.substring(start, end).trim();

        }

        //
        // interval
        //
        if (recurrence.indexOf("INTERVAL") != -1) {
            int start = recurrence.indexOf("INTERVAL=") + "INTERVAL=".length();
            int end   = recurrence.indexOf(";", start);
            value = recurrence.substring(start, end).trim();
            interval = Integer.parseInt(value);

        }

        if ("DAILY".equals(rt)) {
            //
            // daily
            //
            if (recurrence.indexOf(EVERY_WEEKDAY) != -1) {
                recurrenceType = RecurrencePattern.TYPE_WEEKLY;
                dayOfWeekMask = getDayOfWeekMask(EVERY_WEEKDAY);
            } else {
                recurrenceType = RecurrencePattern.TYPE_DAYLY;
            }
        } else if ("WEEKLY".equals(rt)){
            //
            // weekly
            //
            int start = recurrence.indexOf("BYDAY=") + "BYDAY=".length();
            int end   = recurrence.indexOf(";", start);
            value = recurrence.substring(start, end).trim();
            dayOfWeekMask = getDayOfWeekMask(value);
            recurrenceType = RecurrencePattern.TYPE_WEEKLY;
        }  else if ("MONTHLY".equals(rt)){
            if (recurrence.indexOf("BYDAY") != - 1) {
              //
              // montly nt
              //
              int start = recurrence.indexOf("BYDAY=") + "BYDAY=".length();
              int end   = recurrence.indexOf(";", start);
              value = recurrence.substring(start, end).trim();
              instance = Short.parseShort(value.substring(0,value.indexOf(" ")));
              if (instance == -1) {
                  //
                  // last
                  //
                  instance = 5;
              }
              dayOfWeekMask = getDayOfWeekMask(value.substring(value.indexOf(" ") + 1));
              recurrenceType =RecurrencePattern.TYPE_MONTH_NTH;

            } else {
                //
                // montly
                //
                int start = recurrence.indexOf("BYMONTHDAY=") + "BYMONTHDAY=".length();
                int end   = recurrence.indexOf(";", start);
                value = recurrence.substring(start, end).trim();
                dayOfMonth = Short.parseShort(value);
                recurrenceType =RecurrencePattern.TYPE_MONTHLY;
            }

        }  else if ("YEARLY".equals(rt)){

            int start = recurrence.indexOf("BYMONTH=") + "BYMONTH=".length();
            int end  = recurrence.indexOf(";", start);
            value = recurrence.substring(start, end).trim();
            monthOfYear = Short.parseShort(value);

            if (recurrence.indexOf("BYDAY") != - 1) {
              //
              // yearly nt
              //
              start = recurrence.indexOf("BYDAY=") + "BYDAY=".length();
              end   = recurrence.indexOf(";", start);
              value = recurrence.substring(start, end).trim();
              instance = Short.parseShort(value.substring(0,value.indexOf(" ")));
              if (instance == -1) {
                  //
                  // last
                  //
                  instance = 5;
              }
              dayOfWeekMask = getDayOfWeekMask(value.substring(value.indexOf(" ") + 1));

              recurrenceType =RecurrencePattern.TYPE_YEAR_NTH;
           } else {
              //
              // yearly
              //
              start = recurrence.indexOf("BYMONTHDAY=") + "BYMONTHDAY=".length();
              end = recurrence.indexOf(";", start);
              value = recurrence.substring(start, end).trim();
              dayOfMonth = Short.parseShort(value);

              recurrenceType =RecurrencePattern.TYPE_YEARLY;
           }

        }

        switch (recurrenceType) {
            case RecurrencePattern.TYPE_DAYLY:
             if(!noEndDate) {
                rp = RecurrencePattern.getDailyRecurrencePattern(interval        ,
                                                                 startDatePattern,
                                                                 endDatePattern  ,
                                                                 noEndDate);
            } else {
                rp = RecurrencePattern.getDailyRecurrencePattern(interval        ,
                                                                 startDatePattern,
                                                                 noEndDate);
            }
            break;
            case RecurrencePattern.TYPE_WEEKLY:
            if(!noEndDate) {
                rp = RecurrencePattern.getWeeklyRecurrencePattern(interval         ,
                                                                  dayOfWeekMask    ,
                                                                  startDatePattern ,
                                                                  endDatePattern   ,
                                                                  noEndDate);
            } else {
                rp = RecurrencePattern.getWeeklyRecurrencePattern(interval         ,
                                                                  dayOfWeekMask    ,
                                                                  startDatePattern ,
                                                                  noEndDate);
            }
            break;
            case RecurrencePattern.TYPE_MONTHLY:
            if(!noEndDate) {
                rp = RecurrencePattern.getMonthlyRecurrencePattern(interval         ,
                                                                   dayOfMonth       ,
                                                                   startDatePattern ,
                                                                   endDatePattern   ,
                                                                   noEndDate);

            } else {
                rp = RecurrencePattern.getMonthlyRecurrencePattern(interval         ,
                                                                   dayOfMonth       ,
                                                                   startDatePattern ,
                                                                   noEndDate );
            }
            break;
            case RecurrencePattern.TYPE_MONTH_NTH:
            if(!noEndDate) {
                rp = RecurrencePattern.getMonthNthRecurrencePattern(interval         ,
                                                                    dayOfWeekMask    ,
                                                                    instance         ,
                                                                    startDatePattern ,
                                                                    endDatePattern   ,
                                                                    noEndDate);
            } else {
                rp = RecurrencePattern.getMonthNthRecurrencePattern(interval         ,
                                                                    dayOfWeekMask    ,
                                                                    instance         ,
                                                                    startDatePattern ,
                                                                    noEndDate);
            }
            break;
            case RecurrencePattern.TYPE_YEARLY:
            if(!noEndDate) {
                rp = RecurrencePattern.getYearlyRecurrencePattern(dayOfMonth       ,
                                                                  monthOfYear      ,
                                                                  startDatePattern ,
                                                                  endDatePattern   ,
                                                                  noEndDate);

            } else {
                rp = RecurrencePattern.getYearlyRecurrencePattern(dayOfMonth       ,
                                                                  monthOfYear      ,
                                                                  startDatePattern ,
                                                                  noEndDate);
            }
            break;
            case RecurrencePattern.TYPE_YEAR_NTH:
            if(!noEndDate) {
                rp = RecurrencePattern.getYearNthRecurrencePattern(dayOfWeekMask    ,
                                                                   monthOfYear      ,
                                                                   instance         ,
                                                                   startDatePattern ,
                                                                   endDatePattern   ,
                                                                   noEndDate);
            } else {
                rp = RecurrencePattern.getYearNthRecurrencePattern(dayOfWeekMask    ,
                                                                   monthOfYear      ,
                                                                   instance         ,
                                                                   startDatePattern ,
                                                                   noEndDate);
            }

        }

        return rp;

    }

    private String getExchangeRecurrence(RecurrencePattern rp)
    throws Exception {

        String exr = null;

        String type = null;
        String interval = null;
        String endPatternDate = null;

        String dayOfMonth = null;
        String monthOfYear = null;

        boolean noEndDate = false;

        short instance = 0;

        interval = ";INTERVAL=" + rp.getInterval();

        endPatternDate = rp.getEndDatePattern();


        noEndDate = rp.isNoEndDate();

        if (!noEndDate) {
            endPatternDate = ";UNTIL=" + endPatternDate;
        } else {
            endPatternDate = "" ;
        }

        switch (rp.getType()) {
            case RecurrencePattern.TYPE_DAYLY:
               exr = "FREQ=DAILY"     +
                     endPatternDate +
                     interval         +
                     ";WKST=SU"       ;
                break;
            case RecurrencePattern.TYPE_WEEKLY:

                String dayOfWeekMask = null;

                dayOfWeekMask = getDayOfWeekMask(rp.getDayOfWeekMask());

                if (dayOfWeekMask != null && dayOfWeekMask.length() > 0) {
                    dayOfWeekMask = ";BYDAY=" + dayOfWeekMask;
                } else {
                    dayOfWeekMask = "";
                }

                exr = "FREQ=WEEKLY" +
                     endPatternDate +
                     interval       +
                     dayOfWeekMask  +
                     ";WKST=SU"     ;
                break ;
            case RecurrencePattern.TYPE_MONTHLY:

                dayOfMonth = ";BYMONTHDAY=" + rp.getDayOfMonth();

                exr = "FREQ=MONTHLY" +
                      endPatternDate +
                      interval       +
                      dayOfMonth  +
                      ";WKST=SU"     ;

                break;
            case RecurrencePattern.TYPE_MONTH_NTH:

                String instanceDayOfWeekMask = null;

                instance = rp.getInstance();

                dayOfWeekMask = getDayOfWeekMask(rp.getDayOfWeekMask());

                if (dayOfWeekMask == null) {
                     dayOfWeekMask = "";
                }
                instanceDayOfWeekMask = ";BYDAY=" + instance + dayOfWeekMask;

                exr = "FREQ=MONTHLY"         +
                      endPatternDate         +
                      interval               +
                      instanceDayOfWeekMask  +
                      ";WKST=SU"             ;

                break;
            case RecurrencePattern.TYPE_YEARLY:

                 dayOfMonth  = ";BYMONTHDAY=" + rp.getDayOfMonth() ;
                 monthOfYear = ";BYMONTH=" + rp.getMonthOfYear();

                 exr = "FREQ=YEARLY" +
                      endPatternDate +
                      interval       +
                      dayOfMonth     +
                      ";WKST=SU"     ;

                break;

             case RecurrencePattern.TYPE_YEAR_NTH:

                instanceDayOfWeekMask = null;

                instance = rp.getInstance();

                dayOfWeekMask = getDayOfWeekMask(rp.getDayOfWeekMask());

                if (dayOfWeekMask == null) {
                     dayOfWeekMask = "";
                }
                instanceDayOfWeekMask = ";BYDAY=" + instance + dayOfWeekMask;
                monthOfYear = ";BYMONTH=" + rp.getMonthOfYear();

                exr = "FREQ=YEARLY"          +
                      endPatternDate         +
                      interval               +
                      instanceDayOfWeekMask  +
                      monthOfYear            +
                      ";WKST=SU"             ;

                break;


        }

        return exr;


    }

    private short getDayOfWeekMask(String value) {

        short dayOfWeekMask = 0;
        String day = null;

        StringTokenizer st = new StringTokenizer(value, ",");

        while (st.hasMoreTokens())  {

            day = st.nextToken().trim();

            if ("SU".equals(day)) {
                dayOfWeekMask = (short) (dayOfWeekMask + RecurrencePattern.DAY_OF_WEEK_SUNDAY)   ;
            } else if ("MO".equals(day)) {
                dayOfWeekMask = (short) (dayOfWeekMask + RecurrencePattern.DAY_OF_WEEK_MONDAY)   ;
            } else if ("TU".equals(day)) {
                dayOfWeekMask = (short) (dayOfWeekMask + RecurrencePattern.DAY_OF_WEEK_TUESDAY)  ;
            } else if ("WE".equals(day)) {
                dayOfWeekMask = (short) (dayOfWeekMask + RecurrencePattern.DAY_OF_WEEK_WEDNSDAY) ;
            } else if ("TH".equals(day)) {
                dayOfWeekMask = (short) (dayOfWeekMask + RecurrencePattern.DAY_OF_WEEK_THURSDAY) ;
            } else if ("FR".equals(day)) {
                dayOfWeekMask = (short) (dayOfWeekMask + RecurrencePattern.DAY_OF_WEEK_FRIDAY)   ;
            } else if ("SA".equals(day)) {
                dayOfWeekMask = (short) (dayOfWeekMask + RecurrencePattern.DAY_OF_WEEK_SATURDAY) ;
            }

        }

        return dayOfWeekMask;

    }

    private String getDayOfWeekMask(short value) {

        String dayOfWeekMask = "";

        if ((value & RecurrencePattern.DAY_OF_WEEK_SUNDAY) != 0 ) {
            dayOfWeekMask = "SU";
        }

        if ((value & RecurrencePattern.DAY_OF_WEEK_MONDAY) != 0 ) {

            if (dayOfWeekMask.length() > 0) {
               dayOfWeekMask = dayOfWeekMask + ", ";
            }

            dayOfWeekMask = dayOfWeekMask + "MO";
        }

         if ((value & RecurrencePattern.DAY_OF_WEEK_TUESDAY) != 0 ) {

            if (dayOfWeekMask.length() > 0) {
               dayOfWeekMask = dayOfWeekMask + ", ";
            }

            dayOfWeekMask = dayOfWeekMask + "TU";
        }

        if ((value & RecurrencePattern.DAY_OF_WEEK_WEDNSDAY) != 0 ) {

            if (dayOfWeekMask.length() > 0) {
               dayOfWeekMask = dayOfWeekMask + ", ";
            }

            dayOfWeekMask = dayOfWeekMask + "WE";
        }

        if ((value & RecurrencePattern.DAY_OF_WEEK_THURSDAY) != 0 ) {

            if (dayOfWeekMask.length() > 0) {
               dayOfWeekMask = dayOfWeekMask + ", ";
            }

            dayOfWeekMask = dayOfWeekMask + "TH";
        }

        if ((value & RecurrencePattern.DAY_OF_WEEK_FRIDAY) != 0 ) {

            if (dayOfWeekMask.length() > 0) {
               dayOfWeekMask = dayOfWeekMask + ", ";
            }

            dayOfWeekMask = dayOfWeekMask + "FR";
        }

        if ((value & RecurrencePattern.DAY_OF_WEEK_SATURDAY) != 0 ) {

            if (dayOfWeekMask.length() > 0) {
               dayOfWeekMask = dayOfWeekMask + ", ";
            }

            dayOfWeekMask = dayOfWeekMask + "SA";
        }

        return dayOfWeekMask;


    }

}
