/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package sync4j.exchange.xml;

import java.text.MessageFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;

import java.util.ArrayList;
import java.util.Date;
import java.util.StringTokenizer;
import java.util.TimeZone;

/**
 * This class provides commons methods
 * to parser XML msg
 *
 * @author Fabio Maggi @ Funambol
 * @version $Id: XmlParser.java,v 1.19 2005/06/24 09:31:23 fabius Exp $
 **/
public class XmlParser {

    //---------------------------------------------------------------- Constants

    private static final String ANNIVERSARY_FORMAT_PDI    =
            "yyyy-MM-dd"                                            ;

    private static final String ANNIVERSARY_FORMAT_WEBDAV =
            "yyyy-MM-dd'T'HH:mm:ss'.000Z'"                          ;

    private static final String DATE_FORMAT_WEBDAV     =
            "yyyy-MM-dd HH:mm:ss"                                   ;

    private static final String DATE_FORMAT_PDI        =
            "yyyyMMdd'T'HHmmss'Z'"                                  ;

    private static final String DATETIME_FORMAT_CLIENT_UTC =
             "yyyyMMdd'T'HHmmss'Z'"                                 ;

    private static final String DATETIME_FORMAT_CLIENT_NO_UTC =
             "yyyyMMdd'T'HHmmss"                                 ;

    private static final String HEADER_REPLID_START    = "<rid:"    ;

    //---------------------------------------------------------------- Public methods

    /**
     * Make a String[] by tags find with search.
     *
     * @param xmlInput tags about search
     * @param tagContent tag to find
     * @return find tags
     **/
    public static String[] getXMLTag(String[] xmlInput, String tagContent)
    throws XmlParseException {

        ArrayList rows     = null;

        String xmlInputTag = null ;
        String startTag    = null ;
        String endTag      = null ;

        int i = 0;

        startTag = "<"  + tagContent + ">" ;
        endTag   = "</" + tagContent + ">" ;

        rows = new ArrayList();

        try {

            for (int j=0, l = xmlInput.length; j < l; j++) {

                xmlInputTag = xmlInput[j];

                while (xmlInputTag.indexOf(startTag) != -1){
                    rows.add
                        (xmlInputTag.substring(xmlInputTag.indexOf(startTag)         +
                         startTag.length(), xmlInputTag.indexOf(endTag)))            ;
                    xmlInputTag = xmlInputTag.substring(xmlInputTag.indexOf(endTag)  +
                        endTag.length())                                             ;
                    i++;
                }

            }

        } catch (java.lang.StringIndexOutOfBoundsException e) {
            throw new XmlParseException(e.toString() + " : " + tagContent + " TAG");
        }

        return (String[])rows.toArray(new String[i]);

    }

    /**
     * Make a String by value of <i>tag</i>.
     *
     * @param xml xml msg
     * @param tag tag to find
     * @return tag value
     **/
    public static String getXMLTagValue(String xml, String tag)
    throws XmlParseException {

        String msg       = null ;
        String startTag  = null ;
        String endTag    = null ;

        String voidTag   = null ;

        voidTag  = "<"  + tag + "/>" ;

        startTag = "<"  + tag + ">" ;
        endTag   = "</" + tag + ">" ;


        if (xml.indexOf(voidTag) != -1 || xml.indexOf(startTag) == -1) {
            return null;
        }

        try {
            msg = xml.substring(xml.indexOf(startTag)  +
                                startTag.length()      ,
                                xml.indexOf(endTag))   ;
        } catch (java.lang.StringIndexOutOfBoundsException e) {
            throw new XmlParseException(e.toString() + " : " + tag + " TAG");
        }

        return msg;

    }

    /**
     * Make a String by value of <i>tag</i>.
     *
     * @param xml xml msg
     * @param tag tag to find
     * @return tag value
     **/
    public static String getXMLInitTagValue(String xml, String tag)
    throws XmlParseException {

        String msg       = null ;
        String startTag  = null ;
        String endTag    = null ;
        String voidTag   = null ;

        int begin        = 0;
        int init         = 0;

        voidTag  = "<"  + tag + "/>" ;
        startTag = "<"  + tag        ;
        endTag   = "</" + tag + ">"  ;

        if (xml.indexOf(voidTag) != -1 || xml.indexOf(startTag) == -1) {
            return null;
        }

        try {

            begin = xml.indexOf(startTag)                   ;

            init  = xml.indexOf(">", begin) + 1             ;

            //
            // tag void
            //
            if (xml.charAt(init) == '<' && xml.charAt(init + 1) == "/".charAt(0)) {
                return "";
            }

            msg = xml.substring(init, xml.indexOf(endTag))  ;

        } catch (java.lang.StringIndexOutOfBoundsException e) {
            throw new XmlParseException(e.toString() + " : " + tag + " TAG");
        }

        return msg;
    }


   /**
    * Make a href (reource name) from uri webdavtag.
    *
    * @param uri
    * @return id
    * @throws XmlParseException
    **/
   public static String getHrefFromUri(String uri)
    throws XmlParseException {

        String href = null;

        try {

            href = uri.substring(uri.lastIndexOf("/") + 1);

        } catch (java.lang.StringIndexOutOfBoundsException e) {
            throw new XmlParseException(e.toString() + " URI");
        }

        return href;

    }

    /**
     * Make a ruid from a response header.
     *
     * @param response
     * @return ruid
     * @throws XmlParseException
     **/
    public static String getRuidFromResponse(String response)
    throws XmlParseException {

        String ruid = null;

        try {

            int start = response.indexOf(HEADER_REPLID_START)          +
                                         HEADER_REPLID_START.length()  ;

            int end   = response.indexOf(">", start)                   ;

            ruid      = response.substring(start, end)                 ;

        } catch (java.lang.StringIndexOutOfBoundsException e) {
            throw new XmlParseException(e.toString() + " URI");
        }

        return ruid;

    }


    /**
     * Make a date from a webdav tag date to client date
     *
     * @param date webdav tag content
     * @return a Date
     * @throws XmlParseException
     **/
    public static String  toClientDate (Date date)
    throws XmlParseException {

        SimpleDateFormat  formatter    = null ;
        String            value        = null ;

        try {

            if (date != null) {

                formatter = new SimpleDateFormat(DATETIME_FORMAT_CLIENT_UTC);

                return formatter.format(date);

            } else {
                return null;
            }
        } catch (Exception e) {
            throw new XmlParseException(e.toString());
        }

    }

    /**
     * Make a date from a client date to webdav date
     *
     * @param date client date
     * @return a Date
     * @throws XmlParseException
     **/
    public static Date clientDateToDate (String date)
    throws XmlParseException {

        SimpleDateFormat  formatter    = null ;
        Date              timestamp    = null ;

        String            format       = null ;

        try {

            if (date != null) {

                if (date.indexOf("Z") != -1) {
                    format = DATETIME_FORMAT_CLIENT_UTC;
                } else {
                    //
                    // for nokia phone utc bug
                    //
                    format = DATETIME_FORMAT_CLIENT_NO_UTC;
                }

                formatter = new SimpleDateFormat(format);

                timestamp = formatter.parse(date);

                return timestamp;

            } else {
                return null;
            }
        } catch (ParseException e) {
            throw new XmlParseException(e.toString());
        }

    }

    /**
     * Make a date from a webdav tag date
     *
     * @param date webdav tag content
     * @return a Date
     * @throws XmlParseException
     **/
    public static Date webDavTagToDate (String date)
    throws XmlParseException {

        SimpleDateFormat  formatter    = null ;
        String            value        = null ;
        Date              timestamp    = null ;

        try {

            if (date != null) {
                value = date.substring(0,  10) +
                        " "                    +
                        date.substring(11, 19) ;

                formatter = new SimpleDateFormat(DATE_FORMAT_WEBDAV);

                timestamp = formatter.parse(value);
            } else {
                timestamp = null;
            }
        } catch (ParseException e) {
            throw new XmlParseException(e.toString());
        }

        return timestamp;

    }

    /**
     * Make a date from a webdav tag date
     *
     * @param date webdav tag content
     * @return a Date
     * @throws XmlParseException
     **/
    public static String webDavTagToPDIDate (String date)
    throws XmlParseException {

        SimpleDateFormat  formatter    = null ;

        Date              dt           = null ;
        String            pdiDate      = null ;

        if (date != null && date.length() > 0) {

            dt = webDavTagToDate (date);

            formatter = new SimpleDateFormat(DATE_FORMAT_PDI);

            pdiDate = formatter.format(dt);

        } else {
            pdiDate = null;
        }

        return pdiDate;

    }


    /**
     * Make a date from a webdav tag date
     *
     * @param date webdav tag content
     * @return a Date
     * @throws XmlParseException
     **/
    public static Date pdiToDate (String date)
    throws XmlParseException {

        SimpleDateFormat  formatter    = null ;
        Date              dt           = null ;

        try {

            if (date != null) {

                formatter = new SimpleDateFormat(DATE_FORMAT_PDI);

                dt = formatter.parse(date);
            } else {
                dt = null;
            }
        } catch (ParseException e) {
            throw new XmlParseException(e.toString());
        }

        return dt;

    }

    /**
     * Make a date from a webdav tag date
     *
     * @param date webdav tag content
     * @return a Date
     * @throws XmlParseException
     **/
    public static String webDavToPDIAnniversary (String date)
    throws Exception {

        SimpleDateFormat  formatter    = null ;

        Date              dt           = null ;
        String            pdiDate      = null ;

        if (date != null && date.length() > 0) {

            formatter = new SimpleDateFormat(ANNIVERSARY_FORMAT_WEBDAV);

            dt = formatter.parse(date);

            formatter = new SimpleDateFormat(ANNIVERSARY_FORMAT_PDI);

            pdiDate = formatter.format(dt);

        } else {
            pdiDate = null;
        }

        return pdiDate;

    }

    /**
     * Make a date from a webdav tag date
     *
     * @param date webdav tag content
     * @return a Date
     * @throws XmlParseException
     **/
    public static String PDIToWebDavAnniversary (String date)
    throws Exception {

        SimpleDateFormat  formatter    = null ;

        Date              dt           = null ;
        String            pdiDate      = null ;

        if (date != null && date.length() > 0) {

            formatter = new SimpleDateFormat(ANNIVERSARY_FORMAT_PDI);

            dt = formatter.parse(date);

            formatter = new SimpleDateFormat(ANNIVERSARY_FORMAT_WEBDAV);

            pdiDate = formatter.format(dt);

        } else {
            pdiDate = null;
        }

        return pdiDate;

    }

    /**
     * @param value
     * @return <p>value<p> if value is not null, else ""
     **/
    public static String getFieldFromMultipleTag (String value)
    throws Exception {

        String [] fields = null;
        String    field  = null;

        fields = getXMLTag(new String[] {value}, "c:v");


        for (int i = 0, l = fields.length; i < l; i++) {

            if (i == 0) {
                field = fields[0];
            } else {
                field = field + "; " + fields[i];
            }
        }

        return field;

    }

    /**
     * @param value
     * @return <p>value<p> if value is not null, else ""
     **/
    public static String fieldToMultipleTag (String categories) {

        String fields = "" ;

        StringTokenizer st = new StringTokenizer(categories, ";");

        while (st.hasMoreTokens()) {
            fields = fields +   "<x:v>"               +
                                st.nextToken().trim() +
                                "</x:v>\n"            ;
        }

        return fields;
    }

    /**
     * @param value
     * @return <p>value<p> if value is not null, else ""
     **/
    public static String getString (String value) {

        if (value == null) {
            return "";
        } else  {
            return value;
        }

    }


}
